/*
** DGNRESID.g
**
** purpose : provide comprehensive set of diagnostic checks for residuals
**           from estimated univariate time series model
**
** format  : { mean,min,max,med,stdev,mad,skew,pskew,kurt,pkurt,
**             JB,pJB,corr,LB,pLB,LM_sc,pLM_sc,ARCH,pARCH,AIC,BIC,HQ }=
**                 DGNRESID(y,k,m,q,x);
**
** input   : y     : Tx1, vector containing residuals to be diagnozed
**           k     : scalar, number of parameters in estimated model
**           m     : scalar, number of correlations to compute
**           q     : scalar, maximum order of ARCH-statistic to compute
**           x     : Txk, gradient matrix of residuals wrt parameters 
**                        (i.e. in linear model x is regressor matrix)
**
** output  : mean  : scalar, mean of the residual series
**           min   : scalar, minimum of the residual series
**           max   : scalar, maximum of the residual series
**           med   : scalar, median of the residual series
**           stdev : scalar, standard deviation of the residual series
**           mad   : scalar, scaled median absolute deviation
**           skew  : scalar, skewness of the residual series
**           pskew : scalar, probability that skewness is smaller (larger)
**                           than observed if skew is negative (positive)
**           kurt  : scalar, kurtosis of the residual series
**           pkurt : scalar, probability that kurtosis is smaller (larger)
**                           than observed if kurt-3 is negative (positive)
**           JB    : scalar, Jarque-Bera normality test statistic
**           pJB   : scalar, corresponding p-value
**           corr  : (mx1), correlation coefficients
**           LB    : mx1, Ljung-Box test statistic for autocorrelation
**           pLB   : mx1, corresponding p-values
**           LM_sc : mx1, Lagrange Multiplier test statistic for autocorrelation
**           pLM_sc: mx1, corresponding p-values
**           ARCH  : qx1, LM test statistic for ARCH
**           pARCH : qx1, corresponding p-values
**           AIC   : scalar, Akaike Information Criterion
**           BIC   : scalar, Schwarz Information Criterion
**           HQ    : scalar, Hannan-Quinn criterion
**
** globals :
**
** remarks :
**
** authors : Dick van Dijk
**
** date    : First attempt on 20 May 1997
**           Last revision on 20 May 1997
**
*/
proc (22)=DGNRESID(y,k,m,q,gradc);
  local T,mean,min,max,med,stdev,mad,skew,pskew,kurt,pkurt,JB,pJB,corr,
        i,x,z,Tk,LB,pLB,LM_sc,pLM_SC,lagres,aux_m,res,
        yy,myy,vyy,yyreg,b,e,r2,ARCH,pARCH,AIC,BIC,HQ;

T=rows(y);

mean=meanc(y);
min=minc(y);
max=maxc(y);
med=median(y);
mad=1.483.*median(abs(y-med'));
stdev=sqrt(y'y/T);

skew=(sumc(y.^3)./T)./(stdev.^3);
if (skew LT 0);
  pskew=CDFN(skew/sqrt(6/T));
else;
  pskew=CDFNC(skew/sqrt(6/T));
endif;
kurt=(sumc(y.^4)./T)./(stdev.^4);
if (kurt LT 3);
  pkurt=CDFN((kurt-3)/sqrt(24/T));
else;
  pkurt=CDFNC((kurt-3)/sqrt(24/T));
endif;

JB=(T/6)*(skew^2+((kurt-3)^2)/4);
pJB=CDFCHIC(JB,2);

if (m GT 0);
  LM_SC=zeros(m,2);
  pLM_SC=zeros(m,2);
  i=0;
  do until (i==m);
    i=i+1;
    lagres=seqa(1,1,i);
    aux_m=trimr(gradc~rgr(y,lagres),i,0);
    res=trimr(y,i,0);
    {LM_SC[i,1],pLM_SC[i,1],LM_SC[i,2],pLM_SC[i,2]}=CMPLMT(res,cols(gradc),aux_m,i);
  endo;  
  corr=zeros(m,1);
  LB=zeros(m,1);
  pLB=zeros(m,1);

  x=rgr(y,seqa(0,1,m+1));
  corr=((x[.,1]'x[.,2:m+1])./(x[.,1]'x[.,1]))';
  Tk=seqa(T-1,-1,m);
  LB=T*(T+2)*cumsumc((corr.^2)./Tk);
  if (m GT k);
    pLB[k+1:m]=CDFCHIC(LB[k+1:m],seqa(1,1,m-k));
  endif;
else;
  corr=0;
  LB=0;
  pLB=0;
  LM_SC=0;
  pLM_SC=0;
endif;

if (q GT 0);
  yy=y.^2;

  ARCH=zeros(q,1);
  pARCH=zeros(q,1);
  i=0;
  do until (i==q);
    i=i+1;
    myy=meanc(yy[i+1:T]);
    vyy=sumc((yy-myy).^2);
    yyreg=ones(T-i,1)~trimr(rgr(yy,seqa(1,1,i)),i,0);
    b=yy[i+1:T]/yyreg[.,1:i+1];
    e=yy[i+1:T]-yyreg[.,1:i+1]*b;
    r2=1 - e'e/vyy;
    ARCH[i]=T*r2;
    pARCH[i]=CDFCHIC(ARCH[i],i);
  endo;
else;
  ARCH=0;
  pARCH=0;
endif;

AIC=ln(stdev^2)+2*k/T;
BIC=ln(stdev^2)+k*ln(T)/T;
HQ=ln(stdev^2)+2*k*ln(ln(T))/T;

retp(mean,min,max,med,stdev,mad,skew,pskew,kurt,pkurt,JB,pJB,corr,LB,pLB,
     LM_sc,pLM_SC,ARCH,pARCH,AIC,BIC,HQ);
endp;

proc(1)=rgr(x,lg);
  local regr;

regr=shiftr(x',lg,0)';

retp(regr);
endp;

proc(4)=CMPLMT(res_0,k_0,aux_m,k_restr);
  local stdv,stdvzero,cind,T,SSR_0,invaux_m,phi_1,res_1,SSR_1,LM_stc,
        LM_pvc,LM_stf,LM_pvf;

T=rows(res_0);
res_1=1;
SSR_0=res_0'res_0;

invaux_m=invpd(moment(aux_m,0));
if not scalmiss(invaux_m);
  phi_1=invaux_m*(aux_m'res_0);
  res_1=res_0 - aux_m*phi_1;
  SSR_1=res_1'res_1;

  LM_stc=T*(SSR_0 - SSR_1)/SSR_0;
  LM_pvc=CDFCHIC(LM_stc,k_restr);
  LM_stf=((SSR_0 - SSR_1)/k_restr)/(SSR_1/(T-k_restr-k_0));
  LM_pvf=CDFFC(LM_stf,k_restr,T-k_restr-k_0);
else;
  LM_stc=0;
  LM_pvc=0;
  LM_stf=0;
  LM_pvf=0;
endif;

retp(LM_stc,LM_pvc,LM_stf,LM_pvf);
endp;
