/*
** STRTESTS.PRC
** Source file with procedures for computing nonlinearity tests
**
**  Format    Purpose                                                      Line
** ============================================================================
** SSTRTX     Compute LM-type tests against STR nonlinearity                 34
** WSTRTX     Compute heteroskedasticity-robust versions of LM-type         174
**            tests using Wooldridge procedure
** RSTRTX     Compute outlier-robust versions of LM-type tests              308
**            using (G)M-estimator
** SSTRSPA    Compute LM-type tests against STR alternative used in         483
**            specification procedures of Terasvirta (1994),
**            generalization of this rule by Escribano and Jorda [EJ],
**            and the EJ-rule itself
** WSTRSPA    Compute LM-type tests against STR alternative used in         648
**            specification procedures of Terasvirta (1994),
**            generalization of this rule by Escribano and Jorda [EJ],
**            and the EJ-rule itself
**            Robust versions of the tests are computed using procedure 
**            3.1 outlined in Wooldridge (1991)
** RSTRSPA    Compute outlier-robust LM-type tests against STR alternative  798
**            used in specification procedures of Terasvirta (1994),
**            generalization of this rule by Escribano and Jorda [EJ],
**            and the EJ-rule itself
**            M- or GM-estimator is used to estimate models under H0
** DETAUXP    Determine auxiliary regressors                               1089
** CMPLMT     Compute standard LM statistic                                1197
** CMPLMTI    Compute standard LM statistic                                1226
** CMPLMTW    Compute heteroskedasticity-robust LM statistic               1254
*/


/*
** SSTRTX.g
**
** purpose : Compute LM-type tests of Luukkonen et al. (1988) against LSTR
**           and tests of Granger and Terasvirta (1993) and Escribano and
**           Jorda (1998) against ESTR
**
** format  : {LM_1,pv_LM1,LM_2,pv_LM2,LM_3,pv_LM3,LM_3e,pv_LM3e,LM_4,pv_LM4}
**             = SSTRTX(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
**
** input   : dat    : (T x k) matrix, containing data
**           yi     : scalar, index of dependent variable
**           xli    : (k_l x 1), indices of independent variables in
**                    linear part
**           cli    : scalar, index in xli of index of constant if any
**                      If cli=0, no constant in linear model part
**           xnli   : (k_nl x 1), indices of independent variables in
**                    nonlinear part
**           cnli   : scalar, index in xnli of index of constant if any
**                      If cnli=0, no constant in nonlinear model part
**           xtri   : scalar, index of transition variables
**           exc    : 0-1, indicating whether exception has to be made with
**                    respect to inclusion of transition variable as genuine
**                    regressor (relevant if first difference or c.e. is used
**                    as transition variable)
**           res    : either scalar 0 or (T x 1) vector of residuals under
**                    null hypothesis (primarily for use with VECM models)
**
** output  : LM_i   : LM_i statistic, i=1,2,3,3e,4
**           pv_LMi : corresponding p-value, i=1,2,3,3e,4
**
** globals : none
**
** remarks : Both chi-square and F-variants of the test statistics are computed
**
** authors : Dick van Dijk
**
** date    : First attempt on February 3, 1996
**           Last revision on December 30, 1998
*/
proc (10)=SSTRTX(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
  local T,k_l,k_nl,y,xl,xnl,xtr,trl,k_tl,trnl,k_tnl,k_tnlc,nll,xnlnli,xnlnl,
        k_nlnl,phi_0,res_0,SSR_0,aux_m,j,phi_1,res_1,SSR_1,invaux_m,t1,t2,
        LM_1,LM_2,LM_3,LM_3e,LM_4,pv_LM1,pv_LM2,pv_LM3,pv_LM3e,pv_LM4,
        stdv,stdvzero,cind;

trap 1;

/* determine number of observations and number of regressors in the different
   parts of the model */
T=rows(dat);
k_l=rows(xli);
k_nl=rows(xnli);

{xli,cli,xnli,cnli,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,t1,t2}=
    DETAUXP(xli,cli,0,k_l,xnli,cnli,0,k_nl,xtri,exc);

/* Standardize data matrix to zero mean and unit variance */
stdv=stdc(dat);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(dat)),stdvzero);
dat[.,cind]=(dat[.,cind]-meanc(dat[.,cind])')./(stdv[cind]');

y=dat[.,yi];
xl=dat[.,xli];
xnl=dat[.,xnli];
xtr=dat[.,xtri];
if k_nlnl GT 0;
  xnlnl=dat[.,xnlnli];
endif;

LM_1=zeros(2,1); pv_LM1=2*ones(2,1);
LM_2=zeros(2,1); pv_LM2=2*ones(2,1);
LM_3=zeros(2,1); pv_LM3=2*ones(2,1);
LM_4=zeros(2,1); pv_LM4=2*ones(2,1);
LM_3e=zeros(2,1); pv_LM3e=2*ones(2,1);

/* compute covariance matrix and sum of squared residuals
   under the null hypothesis */
if (res==0);
  phi_0=invpd(moment(xl,0))'(xl'*y);
  res_0=y - xl*phi_0;
else;
  res_0=res;
endif;
SSR_0=res_0'res_0;

/* create regressor matrix for auxiliary regression for LM_1 */
aux_m=xl;
/* add nonlinear regressors which are not in linear model part */
if k_nlnl/=0;
  aux_m=aux_m~xnlnl;
endif;
/* add cross-products of transition variable and nonlinear regressors
   plus transition variable if not in linear part */
if (1+k_tl) .LE k_nl;
  aux_m=aux_m~(xnl[.,1+k_tl:k_nl].*xtr);
  /* compute LM_1 and corresponding pvalue */
  {res_1,LM_1,pv_LM1}=CMPLMT(res_0,k_l,aux_m,k_nlnl+k_nl-k_tl);
else;
  LM_1=miss(zeros(2,1),0);
  pv_LM1=miss(zeros(2,1),0);
endif;

/* create regressor matrix for auxiliary regression for LM_3e */
if not k_tnl;
  aux_m=aux_m~(xtr.^2)~(xtr.^3);
else;
  aux_m=aux_m~(xtr.^3);
endif;

/* compute LM_3e and corresponding pvalue */
{res_1,LM_3e,pv_LM3e}=CMPLMT(res_0,k_l,aux_m,k_nlnl+k_nl-k_tl+2-k_tnl);

/* create regressor matrix for auxiliary regression for LM_2 */
aux_m=aux_m[.,1:k_l+k_nlnl+k_nl-k_tl];
/* add square of transition variable if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^2));

/* compute LM_2 and corresponding pvalue */
{res_1,LM_2,pv_LM2}=CMPLMT(res_0,k_l,aux_m,k_nlnl+2*k_nl-k_tl-k_tnlc);

/* create regressor matrix for auxiliary regression for LM_3 */
/* add transition variable cubed if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^3));

/* compute LM_3 and corresponding pvalue */
{res_1,LM_3,pv_LM3}=CMPLMT(res_0,k_l,aux_m,k_nlnl+3*k_nl-k_tl-2*k_tnlc);

/* create regressor matrix for auxiliary regression for LM_4 */
/* add transition variable cubed if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^4));

/* compute LM_4 and corresponding pvalue */
{res_1,LM_4,pv_LM4}=CMPLMT(res_0,k_l,aux_m,k_nlnl+4*k_nl-k_tl-3*k_tnlc);

retp(LM_1,pv_LM1,LM_2,pv_LM2,LM_3,pv_LM3,LM_3e,pv_LM3e,LM_4,pv_LM4);
endp;


/*
** WSTRTX.g
**
** purpose : Compute LM-type tests of Luukkonen et al. (1988) against LSTR
**           and tests of Granger and Terasvirta (1993) and Escribano and
**           Jorda (1998) against ESTR
**           Robust versions of the tests are computed using procedure 3.1
**           outlined in Wooldridge (1991)
**
** format  : {LM_1,pv_LM1,LM_2,pv_LM2,LM_3,pv_LM3,LM_3e,pv_LM3e,LM_4,pv_LM4}
**             = WSTRTX(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
**
** input   : dat    : (T x k) matrix, containing data
**           yi     : scalar, index of dependent variable
**           xli    : (k_l x 1), indices of independent variables in
**                    linear part
**           cli    : scalar, index in xli of index of constant if any
**                      If cli=0, no constant in linear model part
**           xnli   : (k_nl x 1), indices of independent variables in
**                    nonlinear part
**           cnli   : scalar, index in xnli of index of constant if any
**                      If cnli=0, no constant in nonlinear model part
**           xtri   : scalar, index of transition variables
**           exc    : 0-1, indicating whether exception has to be made with
**                    respect to inclusion of transition variable as genuine
**                    regressor (relevant if first difference or c.e. is used
**                    as transition variable)
**           res    : either scalar 0 or (T x 1) vector of residuals under
**                    null hypothesis (primarily for use with VECM models)
**
** output  : LM_i   : LM_i statistic, i=1,2,3,3e,4
**           pv_LMi : corresponding p-value, i=1,2,3,3e,4
**
** globals : none
**
** remarks : Both chi-square and F-variants of the test statistics are computed
**
** authors : Dick van Dijk
**
** date    : First attempt on February 3, 1996
**           Last revision on January 1, 1999
*/
proc (10)=WSTRTX(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
  local T,k_l,k_nl,y,xl,xnl,xtr,trl,k_tl,trnl,k_tnl,k_tnlc,nll,xnlnli,xnlnl,
        k_nlnl,phi_0,res_0,SSR_0,aux_m,j,phi_1,res_1,SSR_1,invaux_m,t1,t2,
        LM_1,LM_2,LM_3,LM_3e,LM_4,pv_LM1,pv_LM2,pv_LM3,pv_LM3e,pv_LM4,
        stdv,stdvzero,cind;

trap 1;

/* determine number of observations and number of regressors in the different
   parts of the model */
T=rows(dat);
k_l=rows(xli);
k_nl=rows(xnli);

{xli,cli,xnli,cnli,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,t1,t2}=
    DETAUXP(xli,cli,0,k_l,xnli,cnli,0,k_nl,xtri,exc);

/* Standardize data matrix to zero mean and unit variance */
stdv=stdc(dat);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(dat)),stdvzero);
dat[.,cind]=(dat[.,cind]-meanc(dat[.,cind])')./(stdv[cind]');

y=dat[.,yi];
xl=dat[.,xli];
xnl=dat[.,xnli];
xtr=dat[.,xtri];
if k_nlnl GT 0;
  xnlnl=dat[.,xnlnli];
endif;

LM_1=zeros(2,1); pv_LM1=2*ones(2,1);
LM_2=zeros(2,1); pv_LM2=2*ones(2,1);
LM_3=zeros(2,1); pv_LM3=2*ones(2,1);
LM_4=zeros(2,1); pv_LM4=2*ones(2,1);
LM_3e=zeros(2,1); pv_LM3e=2*ones(2,1);

/* create regressor matrix for auxiliary regression for LM_1 */
aux_m=xl;
/* add nonlinear regressors which are not in linear model part */
if k_nlnl/=0;
  aux_m=aux_m~xnlnl;
endif;
/* add cross-products of transition variable and nonlinear regressors
   plus transition variable if not in linear part */
if (1+k_tl) .LE k_nl;
  aux_m=aux_m~(xnl[.,1+k_tl:k_nl].*xtr);

  /* compute LM_1 and corresponding pvalue */
  {LM_1,pv_LM1}=CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+k_nl-k_tl]);
endif;

/* create regressor matrix for auxiliary regression for LM_3e */
if not k_tnl;
  aux_m=aux_m~(xtr.^2)~(xtr.^3);
else;
  aux_m=aux_m~(xtr.^3);
endif;

/* compute LM_3e and corresponding pvalue */
{LM_3e,pv_LM3e}=
   CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+k_nl-k_tl+2-k_tnl]);

/* create regressor matrix for auxiliary regression for LM_2 */
aux_m=aux_m[.,1:k_l+k_nlnl+k_nl-k_tl];
/* add square of transition variable if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^2));

/* compute LM_2 and corresponding pvalue */
{LM_2,pv_LM2}=
   CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc]);

/* create regressor matrix for auxiliary regression for LM_3 */
/* add transition variable cubed if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^3));

/* compute LM_3 and corresponding pvalue */
{LM_3,pv_LM3}=
   CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc]);

/* create regressor matrix for auxiliary regression for LM_4 */
/* add transition variable cubed if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^4));

/* compute LM_4 and corresponding pvalue */
{LM_4,pv_LM4}=
   CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+4*k_nl-k_tl-3*k_tnlc]);

retp(LM_1,pv_LM1,LM_2,pv_LM2,LM_3,pv_LM3,LM_3e,pv_LM3e,LM_4,pv_LM4);
endp;


/*
** RSTRTX.g
**
** purpose : Compute LM-type tests of Luukkonen et al. (1988) against LSTR
**           and tests of Granger and Tersvirta (1993) and Escribano and
**           Jorda (1998) against ESTR
**
**           An HBP-(G)M estimator is used to estimate the parameters of the
**           linear model under the null, using the psi_pol function to
**           compute residual weights
**
** format  : {LM_1,pv_LM1,LM_2,pv_LM2,LM_3,pv_LM3,LM_3e,pv_LM3e,LM_4,pv_LM4,weightse}
**             = RSTRTX(dat,yi,xli,cli,detli,xnli,cnli,detnli,xtri,exc,res,GM);
**
** input   : dat    : (T x k) matrix, containing data
**           yi     : scalar, index of dependent variable
**           xli    : (k_l x 1), indices of independent variables in
**                      linear part
**           cli    : scalar, index in xli of index of constant if any
**                      If cli=0, no constant in linear model part
**           detli  : (k_dlx1), indices of deterministic regressors in linear
**                    part of the model, e.g., (seasonal) dummies
**           xnli   : (k_nl x 1), indices of independent variables in
**                    nonlinear part
**           cnli   : scalar, index in xnli of index of constant if any
**                      If cnli=0, no constant in nonlinear model part
**           detnli : (k_dnlx1), indices of deterministic regressors in
**                    nonlinear part of the model, e.g., (seasonal) dummies
**           xtri   : scalar, index of transition variables
**           exc    : 0-1, indicating whether exception has to be made with
**                    respect to inclusion of transition variable as genuine
**                    regressor (relevant if first difference or c.e. is used
**                    as transition variable)
**           res    : either scalar 0 or (T x 1) vector of residuals under
**                    null hypothesis (primarily for use with VECM models)
**           GM     : 0-1, indicating whether M- (0) or GM-estimator (1) is to
**                    be used
**
** output  : LM_i   : LM_i statistic, i=1,2,3,3e,4
**           pv_LMi : corresponding p-value, i=1,2,3,3e,4
**           weightse : (Tx1) vector, containing weights assigned to
**                       observations by the HBP-(G)M estimation procedure
**
** globals : none
**
** remarks : Both chi-square and F-variants of the test statistics are computed
**
** authors : Dick van Dijk
**
** date    : First attempt on February 3, 1996
**           Last revision on December 30, 1998
*/
proc (11)=RSTRTX(dat,yi,xli,cli,detli,xnli,cnli,detnli,xtri,exc,res,GM);
  local maxit,tol,T,k_l,k_nl,y,xl,xnl,xtr,trl,k_tl,trnl,k_tnl,k_tnlc,nll,
        xnlnli,xnlnl,k_nlnl,phi_0,weightsx,weightse,res_0,SSR_0,sigma,
        psi_0,ifail,aux_m,invaux_m,phi_1,res_1,SSR_1,deteri,
        LM_1,LM_2,LM_3,LM_3e,LM_4,pv_LM1,pv_LM2,pv_LM3,pv_LM3e,pv_LM4,
        stdv,stdvzero,cind;

trap 1;
maxit=1;
tol=1E-3;

/* determine number of observations and number of regressors in the different
   parts of the model */
T=rows(dat);
k_l=rows(xli);
k_nl=rows(xnli);

{xli,cli,xnli,cnli,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,detli,detnli}=
    DETAUXP(xli,cli,detli,k_l,xnli,cnli,detnli,k_nl,xtri,exc);

/* Standardize data matrix to zero mean and unit variance */
stdv=stdc(dat);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(dat)),stdvzero);
dat[.,cind]=(dat[.,cind]-meanc(dat[.,cind])')./(stdv[cind]');

y=dat[.,yi];
xl=dat[.,xli];
xnl=dat[.,xnli];
xtr=dat[.,xtri];
if k_nlnl GT 0;
  xnlnl=dat[.,xnlnli];
endif;
deteri={};
if (detli /=0);
  deteri=deteri|detli;
endif;
if (detnli /=0);
  deteri=deteri|(detnli+k_l);
endif;
if scalmiss(deteri);
  deteri=0;
endif;

LM_1=zeros(2,1); pv_LM1=2*ones(2,1);
LM_2=zeros(2,1); pv_LM2=2*ones(2,1);
LM_3=zeros(2,1); pv_LM3=2*ones(2,1);
LM_3e=zeros(2,1); pv_LM3e=2*ones(2,1);
LM_4=zeros(2,1); pv_LM4=2*ones(2,1);

/* compute sum of squared residuals under the null hypothesis */
if (res==0);
  phi_0=LMS(y,xl);
  if (not GM);
    {phi_0,weightse,ifail}=M_EST(y,xl,phi_0,&psi_pol,maxit,tol);
    weightsx=1;
  else;
    {phi_0,weightsx,weightse,ifail}=
        GM_EST(y,xl,phi_0,&psi_polx,&psi_pol,maxit,tol,deteri);
  endif;
  res_0=y - xl*phi_0;
else;
  res_0=res;
endif;
sigma=madc(res_0);

/* compute psi_function, SSR_0 and matrix with regressor weigths */
psi_0=weightse.*(res_0./(sigma.*(weightsx+(1e-5.*(weightsx.<1e-5)))));
SSR_0=psi_0'psi_0;

/* create regressor matrix for auxiliary regression for LM_1 */
aux_m=xl;
/* add nonlinear regressors which are not in linear model part */
if k_nlnl/=0;
  aux_m=aux_m~xnlnl;
endif;
/* add cross-products of transition variable and nonlinear regressors
   plus transition variable if not in linear part */
if (1+k_tl) .LE k_nl;
  aux_m=aux_m~(xnl[.,1+k_tl:k_nl].*xtr);
  /* compute LM_L1 and corresponding pvalue */
  {res_1,LM_1,pv_LM1}=CMPLMT(psi_0,k_l,weightsx.*aux_m,k_nlnl+k_nl-k_tl);
endif;

/* create regressor matrix for auxiliary regression for LM_3e */
if not k_tnl;
  aux_m=aux_m~(xtr.^2)~(xtr.^3);
else;
  aux_m=aux_m~(xtr.^3);
endif;

/* compute LM_3e and corresponding pvalue */
{res_1,LM_3e,pv_LM3e}=
   CMPLMT(psi_0,k_l,weightsx.*aux_m,k_nlnl+k_nl-k_tl+2-k_tnl);

/* create regressor matrix for auxiliary regression for LM_2 */
aux_m=aux_m[.,1:k_l+k_nlnl+k_nl-k_tl];
/* add square of transition variable if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^2));

/* compute LM_2 and corresponding pvalue */
{res_1,LM_2,pv_LM2}=CMPLMT(psi_0,k_l,weightsx.*aux_m,k_nlnl+2*k_nl-k_tl-k_tnlc);

/* create regressor matrix for auxiliary regression for LM_3 */
/* add transition variable cubed if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^3));

/* compute LM_3 and corresponding pvalue */
{res_1,LM_3,pv_LM3}=
   CMPLMT(psi_0,k_l,weightsx.*aux_m,k_nlnl+3*k_nl-k_tl-2*k_tnlc);

/* create regressor matrix for auxiliary regression for LM_4 */
/* add transition variable cubed if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^4));

/* compute LM_4 and corresponding pvalue */
{res_1,LM_4,pv_LM4}=
   CMPLMT(psi_0,k_l,weightsx.*aux_m,k_nlnl+4*k_nl-k_tl-3*k_tnlc);

retp(LM_1,pv_LM1,LM_2,pv_LM2,LM_3,pv_LM3,LM_3e,pv_LM3e,LM_4,pv_LM4,weightse);
endp;


/*
** SSTRSPA.g
**
** purpose : compute  LM-type tests against STR alternative used in
**           specification procedures of Terasvirta (1994), the generalization
**           of this rule by Escribano and Jorda [EJ], and the EJ-rule itself
**
** format  : {LM_3,pval_3,LM_4,pval_4,LM_H4,pval_H4,LM_H3,pval_H3,LM_H2,
**            pval_H2,LM_H1,pval_H1,LM_HL,pval_HL,LM_HE,pval_HE}
**            = SSTRSPA(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
**
** input   : dat    : (T x k) matrix, containing data
**           yi     : scalar, index of dependent variable
**           xli    : (k_l x 1), indices of independent variables in
**                    linear part
**           cli    : scalar, index in xli of index of constant if any
**                      If cli=0, no constant in linear model part
**           xnli   : (k_nl x 1), indices of independent variables in
**                    nonlinear part
**           cnli   : scalar, index in xnli of index of constant if any
**                      If cnli=0, no constant in nonlinear model part
**           xtri   : scalar, index of transition variables
**           exc    : 0-1, indicating whether exception has to be made with
**                    respect to inclusion of transition variable as genuine
**                    regressor (relevant if first difference or c.e. is used
**                    as transition variable)
**           res    : either scalar 0 or (T x 1) vector of residuals under
**                    null hypothesis (primarily for use with VECM models)
**
** output  : LM_3   : LM-type test against STR alternative
**           LM_4   : LM-type test against STR alternative
**           pval_i : corresponding p-value, i=3,4
**           LM_Hi  : LM-statistic to test H_0i: coefficients of terms of order
**                    (i+1) are equal to zero (conditional on "higher order"
**                    null hypothesis being true, i=1,...,4, or: coefficients
**                    of logistic or exponential terms are equal to zero, i=L,E
**           pval_Hi: corresponding p-value, i=1,2,3,4,L,E
**
** globals : none
**
** remarks : Both chi-square and F-variants of test statistics are computed
**
** authors : Dick van Dijk
**
** date    : First attempt on April 30, 1996
**           Last revision on December 31, 1998
*/
proc (16)=SSTRSPA(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
  local T,k_l,k_nl,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,y,xl,xnl,xtr,xnlnl,
        phi_0,res_0,SSR_0,t1,t2,
        aux_m,res_1,SSR_LM4,SSR_LM3,SSR_L,SSR_E,SSR_3,SSR_2,
        LM_4,LM_HL,LM_HE,LM_3,LM_H4,LM_H3,LM_H2,LM_H1,pval_4,pval_HL,
        pval_HE,pval_3,pval_H4,pval_H3,pval_H2,pval_H1,logi,expi,stdv,stdvzero,cind;

trap 1;

/* determine number of observations and number of regressors in the different
   parts of the model (apart from the intercept) */
T=rows(dat);
k_l=rows(xli);
k_nl=rows(xnli);

{xli,cli,xnli,cnli,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,t1,t2}=
  DETAUXP(xli,cli,0,k_l,xnli,cnli,0,k_nl,xtri,exc);

/* Standardize data matrix to zero mean and unit variance */
stdv=stdc(dat);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(dat)),stdvzero);
dat[.,cind]=(dat[.,cind]-meanc(dat[.,cind])')./(stdv[cind]');

y=dat[.,yi];
xl=dat[.,xli];
xnl=dat[.,xnli];
xtr=dat[.,xtri];
if k_nlnl GT 0;
  xnlnl=dat[.,xnlnli];
endif;

LM_H1=zeros(2,1); pval_H1=2*ones(2,1);
LM_H2=zeros(2,1); pval_H2=2*ones(2,1);
LM_H3=zeros(2,1); pval_H3=2*ones(2,1);
LM_H4=zeros(2,1); pval_H4=2*ones(2,1);
LM_HL=zeros(2,1); pval_HL=2*ones(2,1);
LM_HE=zeros(2,1); pval_HE=2*ones(2,1);
LM_3=zeros(2,1); pval_3=2*ones(2,1);
LM_4=zeros(2,1); pval_4=2*ones(2,1);

/* compute sum of squared residuals under the null hypothesis */
if (res==0);
  phi_0=invpd(moment(xl,0))'(xl'*y);
  res_0=y - xl*phi_0;
else;
  res_0=res;
endif;
SSR_0=res_0'res_0;

/* create regressor matrix for auxiliary regression */
aux_m=xl;
/* add nonlinear regressors which are not in linear model part */
if k_nlnl/=0;
  aux_m=aux_m~xnlnl;
endif;
/* add cross-products of transition variable and nonlinear regressors
   plus transition variable if not in linear part */
if (1+k_tnl) .LE k_nl;
  aux_m=aux_m~(xnl[.,1+k_tl:k_nl].*xtr);
endif;

/* add squares and cube of transition variable if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^2));
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^3));
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^4));

/* compute LM_4 and corresponding pvalue */
{res_1,LM_4,pval_4}=CMPLMT(res_0,k_l,aux_m,k_nlnl+4*k_nl-k_tl-3*k_tnlc);
SSR_LM4=res_1'res_1;

/* compute LM_HL and corresponding pvalue */
logi=seqa(1,1,k_l+k_nlnl)|seqa(k_l+k_nlnl+k_nl-k_tl+1,1,k_nl-k_tnlc)|
     seqa(k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc+1,1,k_nl-k_tnlc);
{res_1,LM_HL,pval_HL}=CMPLMTI(res_0,k_l+k_nlnl+2*k_nl-2*k_tnlc,
        aux_m[.,logi],2*k_nl-k_tl-k_tnlc,SSR_LM4);

/* compute LM_HE and corresponding pvalue */
expi=seqa(1,1,k_l+k_nlnl+k_nl-k_tl)|
     seqa(k_l+k_nlnl+2*k_nl-k_tl-k_tnlc+1,1,k_nl-k_tnlc);
{res_1,LM_HE,pval_HE}=CMPLMTI(res_0,k_l+k_nlnl+2*k_nl-k_tl-k_tnlc,
    aux_m[.,expi],2*k_nl-2*k_tnlc,SSR_LM4);

/* compute LM_3 and corresponding pvalue */
{res_1,LM_3,pval_3}=
   CMPLMT(res_0,k_l,aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc],
           k_nlnl+3*k_nl-k_tl-2*k_tnlc);
SSR_LM3=res_1'res_1;

/* compute LM_H4 and corresponding pvalue */
LM_H4[1]=T*(SSR_LM3-SSR_LM4)/SSR_LM3;
pval_H4[1]=CDFCHIC(LM_H4[1],k_nl-k_tnlc);
LM_H4[2]=((SSR_LM3 - SSR_LM4)/(k_nl-k_tnlc))/
               (SSR_LM4/(T-(k_nlnl+4*k_nl-k_tl-3*k_tnlc+k_l)));
pval_H4[2]=CDFFC(LM_H4[2],k_nl-k_tnlc,T-(k_nlnl+4*k_nl-k_tl-3*k_tnlc+k_l));

/* compute LM_H3 and corresponding pvalue */
{res_1,LM_H3,pval_H3}=CMPLMTI(res_0,k_l+k_nlnl+2*k_nl-k_tl-k_tnlc,
    aux_m[.,1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc],k_nl-k_tnlc,SSR_LM3);
SSR_3=res_1'res_1;

/* compute LM_H2 and corresponding pvalue */
{res_1,LM_H2,pval_H2}=CMPLMTI(res_0,k_l+k_nlnl+k_nl-k_tl,
    aux_m[.,1:k_l+k_nlnl+k_nl-k_tl],k_nl-k_tnlc,SSR_3);
SSR_2=res_1'res_1;

/* compute LM_H1 and corresponding pvalue */
LM_H1[1]=T*(SSR_0-SSR_2)/SSR_0;
pval_H1[1]=CDFCHIC(LM_H1[1],k_nlnl+k_nl-k_tl);
LM_H1[2]=((SSR_0 - SSR_2)/(k_nlnl+k_nl-k_tl))/
               (SSR_2/(T-(k_nlnl+k_nl-k_tl+k_l)));
pval_H1[2]=CDFFC(LM_H1[2],k_nlnl+k_nl-k_tl,T-(k_nlnl+k_nl-k_tl+k_l));

retp(LM_3,pval_3,LM_4,pval_4,LM_H4,pval_H4,LM_H3,pval_H3,LM_H2,pval_H2,
     LM_H1,pval_H1,LM_HL,pval_HL,LM_HE,pval_HE);
endp;


/*
** WSTRSPA.g
**
** purpose : compute LM-type tests against STR alternative used in
**           specification procedures of Terasvirta (1994), the generalization
**           of this rule by Escribano and Jorda [EJ], and the EJ-rule itself
**           Robust versions of the tests are computed using procedure 3.1
**           outlined in Wooldridge (1991)
**
** format  : {LM_3,pval_3,LM_4,pval_4,LM_H4,pval_H4,LM_H3,pval_H3,LM_H2,
**            pval_H2,LM_H1,pval_H1,LM_HL,pval_HL,LM_HE,pval_HE}
**            = WSTRSPA(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
**
** input   : dat    : (T x k) matrix, containing data
**           yi     : scalar, index of dependent variable
**           xli    : (k_l x 1), indices of independent variables in
**                     linear part
**           cli    : scalar, index in xli of index of constant if any
**                      If cli=0, no constant in linear model part
**           xnli   : (k_nl x 1), indices of independent variables in
**                    nonlinear part
**           cnli   : scalar, index in xnli of index of constant if any
**                      If cnli=0, no constant in nonlinear model part
**           xtri   : scalar, index of transition variables
**           exc    : 0-1, indicating whether exception has to be made with
**                    respect to inclusion of transition variable as genuine
**                    regressor (relevant if first difference or c.e. is used
**                    as transition variable)
**           res    : either scalar 0 or (T x 1) vector of residuals under
**                    null hypothesis (primarily for use with VECM models)
**
** output  : LM_3   : LM-type test against STR alternative
**           LM_4   : LM-type test against STR alternative
**           pval_i : corresponding p-value, i=3,4
**           LM_Hi  : LM-statistic to test H_0i: coefficients of terms of order
**                    (i+1) are equal to zero (conditional on "higher order"
**                    null hypothesis being true, i=1,...,4, or: coefficients
**                    of logistic or exponential terms are equal to zero, i=L,E
**           pval_Hi: corresponding p-value, i=1,2,3,4,L,E
**
** globals : none
**
** remarks : Both chi-square and F-variants of tests are computed
**
** authors : Dick van Dijk
**
** date    : First attempt on April 30,    1996
**           Last revision on December 31, 1998
*/
proc (16)=WSTRSPA(dat,yi,xli,cli,xnli,cnli,xtri,exc,res);
  local T,k_l,k_nl,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,y,xl,xnl,xtr,xnlnl,
        phi_0,res_0,SSR_0,t1,t2,aux_m,LM_4,LM_HL,LM_HE,LM_3,LM_H4,
		LM_H3,LM_H2,LM_H1,pval_4,pval_HL,pval_HE,pval_3,pval_H4,pval_H3,
		pval_H2,pval_H1,logi,expi,stdv,stdvzero,cind;

trap 1;

/* determine number of observations and number of regressors in the different
   parts of the model (apart from the intercept) */
T=rows(dat);
k_l=rows(xli);
k_nl=rows(xnli);

{xli,cli,xnli,cnli,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,t1,t2}=
  DETAUXP(xli,cli,0,k_l,xnli,cnli,0,k_nl,xtri,exc);

/* Standardize data matrix to zero mean and unit variance */
stdv=stdc(dat);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(dat)),stdvzero);
dat[.,cind]=(dat[.,cind]-meanc(dat[.,cind])')./(stdv[cind]');

y=dat[.,yi];
xl=dat[.,xli];
xnl=dat[.,xnli];
xtr=dat[.,xtri];
if k_nlnl GT 0;
  xnlnl=dat[.,xnlnli];
endif;

LM_H1=zeros(2,1); pval_H1=2*ones(2,1);
LM_H2=zeros(2,1); pval_H2=2*ones(2,1);
LM_H3=zeros(2,1); pval_H3=2*ones(2,1);
LM_H4=zeros(2,1); pval_H4=2*ones(2,1);
LM_HL=zeros(2,1); pval_HL=2*ones(2,1);
LM_HE=zeros(2,1); pval_HE=2*ones(2,1);
LM_3=zeros(2,1); pval_3=2*ones(2,1);
LM_4=zeros(2,1); pval_4=2*ones(2,1);

/* create regressor matrix for auxiliary regression */
aux_m=xl;
/* add nonlinear regressors which are not in linear model part */
if k_nlnl/=0;
  aux_m=aux_m~xnlnl;
endif;
/* add cross-products of transition variable and nonlinear regressors
   plus transition variable if not in linear part */
if (1+k_tnl) .LE k_nl;
  aux_m=aux_m~(xnl[.,1+k_tl:k_nl].*xtr);
endif;

/* add squares and cube of transition variable if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^2));
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^3));
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^4));

/* compute LM_3 and LM_4 and corresponding pvalues */
{LM_3,pval_3}=
   CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc]);
{LM_4,pval_4}=
   CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+4*k_nl-k_tl-3*k_tnlc]);

/* compute LM_HL and corresponding pvalue */
logi=seqa(1,1,k_l+k_nlnl)|seqa(k_l+k_nlnl+k_nl-k_tl+1,1,k_nl-k_tnlc)|
     seqa(k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc+1,1,k_nl-k_tnlc);
if (k_tl .LT k_tl);
  expi=seqa(k_l+k_nlnl+1,1,k_nl-k_tl)|
        seqa(k_l+k_nlnl+2*k_nl-k_tl-k_tnlc+1,1,k_nl-k_tnlc);
else;
  expi=seqa(k_l+k_nlnl+2*k_nl-k_tl-k_tnlc+1,1,k_nl-k_tnlc);
endif;
{LM_HL,pval_HL}=CMPLMTW(y,aux_m[.,logi],aux_m[.,expi]);

/* compute LM_HE and corresponding pvalue */
logi=seqa(k_l+k_nlnl+k_nl-k_tl+1,1,k_nl-k_tnlc)|
     seqa(k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc+1,1,k_nl-k_tnlc);
expi=seqa(1,1,k_l+k_nlnl+k_nl-k_tl)|
     seqa(k_l+k_nlnl+2*k_nl-k_tl-k_tnlc+1,1,k_nl-k_tnlc);
{LM_HE,pval_HE}=CMPLMTW(y,aux_m[.,expi],aux_m[.,logi]);

/* compute LM_H4 and corresponding pvalue */
{LM_H4,pval_H4}=CMPLMTW(y,aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc],
    aux_m[.,k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc+1:k_l+k_nlnl+4*k_nl-k_tl-3*k_tnlc]);

/* compute LM_H3 and corresponding pvalue */
{LM_H3,pval_H3}=CMPLMTW(y,aux_m[.,1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc],
    aux_m[.,k_l+k_nlnl+2*k_nl-k_tl-k_tnlc+1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc]);

/* compute LM_H2 and corresponding pvalue */
{LM_H2,pval_H2}=CMPLMTW(y,aux_m[.,1:k_l+k_nlnl+k_nl-k_tl],
    aux_m[.,k_l+k_nlnl+k_nl-k_tl+1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc]);

/* compute LM_H1 and corresponding pvalue */
{LM_H1,pval_H1}=CMPLMTW(y,aux_m[.,1:k_l],aux_m[.,k_l+1:k_l+k_nlnl+k_nl-k_tl]);

retp(LM_3,pval_3,LM_4,pval_4,LM_H4,pval_H4,LM_H3,pval_H3,
     LM_H2,pval_H2,LM_H1,pval_H1,LM_HL,pval_HL,LM_HE,pval_HE);
endp;


/*
** RSTRSPA.g
**
** purpose : compute outlier-robust LM-type tests against STR alternative used in
**           specification procedures of Terasvirta (1994), the generalization
**           of this rule by Escribano and Jorda [EJ], and the EJ-rule itself
**
** format  : {LM_3,pval_3,LM_4,pval_4,LM_H4,pval_H4,LM_H3,pval_H3,LM_H2,
**            pval_H2,LM_H1,pval_H1,LM_HL,pval_HL,LM_HE,pval_HE,weights}
**            = RSTRSPA(dat,yi,xli,cli,detli,xnli,cnli,detnli,xtri,exc,res,GM);
**
** input   : dat    : (T x k) matrix, containing data
**           yi     : scalar, index of dependent variable
**           xli    : (k_l x 1), indices of independent variables in
**                      linear part
**           cli    : scalar, index in xli of index of constant if any
**                      If cli=0, no constant in linear model part
**           detli  : (k_dlx1), indices of deterministic regressors in linear
**                    part of the model, e.g., (seasonal) dummies
**           xnli   : (k_nl x 1), indices of independent variables in
**                    nonlinear part
**           cnli   : scalar, index in xnli of index of constant if any
**                      If cnli=0, no constant in nonlinear model part
**           detnli : (k_dnlx1), indices of deterministic regressors in
**                    nonlinear part of the model, e.g., (seasonal) dummies
**           xtri   : scalar, index of transition variables
**           exc    : 0-1, indicating whether exception has to be made with
**                    respect to inclusion of transition variable as genuine
**                    regressor (relevant if first difference or c.e. is used
**                    as transition variable)
**           res    : either scalar 0 or (T x 1) vector of residuals under
**                    null hypothesis (primarily for use with VECM models)
**           GM     : 0-1, indicating whether M- (0) or GM-estimator (1) is to
**                    be used
**
** output  : LM_3   : LM-type test against STR alternative
**           LM_4   : LM-type test against STR alternative
**           pval_i : corresponding p-value, i=3,4
**           LM_Hi  : LM-statistic to test H_0i: coefficients of terms of order
**                    (i+1) are equal to zero (conditional on "higher order"
**                    null hypothesis being true, i=1,...,4, or: coefficients
**                    of logistic or exponential terms are equal to zero, i=L,E
**           pval_Hi: corresponding p-value, i=1,2,3,4,L,E
**           weightse : (Tx1) vector, containing weights assigned to
**                      observations by the HBP-GM estimation procedure of the
**                      linear model
**
** globals : none
**
** remarks :
**
** authors : Dick van Dijk
**
** date    : First attempt on April 30,    1996
**           Last revision on December 31, 1998
*/
proc (17)=RSTRSPA(dat,yi,xli,cli,detli,xnli,cnli,detnli,xtri,exc,res,GM);
  local maxit,tol,T,k_l,k_nl,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,y,xl,xnl,xtr,
        xnlnl,LM_4,LM_HL,LM_HE,LM_3,LM_H4,LM_H3,LM_H2,LM_H1,pval_4,
        pval_3,pval_HL,pval_HE,pval_H4,pval_H3,pval_H2,pval_H1,logi,expi,
        phi_0,weightse,weightsx,ifail,res_0,sigma,psi_0,SSR_0,aux_m,res_1,
        deteri,stdv,stdvzero,cind;

trap 1;
maxit=1;
tol=1E-3;

/* determine number of observations and number of regressors in the different
   parts of the model */
T=rows(dat);
k_l=rows(xli);
k_nl=rows(xnli);

{xli,cli,xnli,cnli,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,detli,detnli}=
  DETAUXP(xli,cli,detli,k_l,xnli,cnli,detnli,k_nl,xtri,exc);

/* Standardize data matrix to zero mean and unit variance */
stdv=stdc(dat);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(dat)),stdvzero);
dat[.,cind]=(dat[.,cind]-meanc(dat[.,cind])')./(stdv[cind]');

y=dat[.,yi];
xl=dat[.,xli];
xnl=dat[.,xnli];
xtr=dat[.,xtri];
if k_nlnl GT 0;
  xnlnl=dat[.,xnlnli];
endif;
deteri={};
if (detli /=0);
  deteri=deteri|detli;
endif;
if (detnli /=0);
  deteri=deteri|(k_l+detnli);
endif;
if scalmiss(deteri);
  deteri=0;
endif;

LM_H1=zeros(2,1); pval_H1=2*ones(2,1);
LM_H2=zeros(2,1); pval_H2=2*ones(2,1);
LM_H3=zeros(2,1); pval_H3=2*ones(2,1);
LM_H4=zeros(2,1); pval_H4=2*ones(2,1);
LM_HL=zeros(2,1); pval_HL=2*ones(2,1);
LM_HE=zeros(2,1); pval_HE=2*ones(2,1);
LM_3=zeros(2,1); pval_3=2*ones(2,1);
LM_4=zeros(2,1); pval_4=2*ones(2,1);

/* create regressor matrix for auxiliary regression */
aux_m=xl;
/* add nonlinear regressors which are not in linear model part */
if k_nlnl/=0;
  aux_m=aux_m~xnlnl;
endif;
/* add cross-products of transition variable and nonlinear regressors
   plus transition variable if not in linear part */
if (1+k_tl) .LE k_nl;
  aux_m=aux_m~(xnl[.,1+k_tl:k_nl].*xtr);
endif;

/* add squares and cube of transition variable if not in nonlinear part */
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^2));
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^3));
aux_m=aux_m~(xnl[.,1+k_tnlc:k_nl].*(xtr.^4));

/* compute sum of squared residuals under the null hypothesis */
if (res==0);
  phi_0=LMS(y,xl);
  if (not GM);
    {phi_0,weightse,ifail}=M_EST(y,xl,phi_0,&psi_pol,maxit,tol);
    weightsx=1;
  else;
    {phi_0,weightsx,weightse,ifail}=
        GM_EST(y,xl,phi_0,&psi_polx,&psi_pol,maxit,tol,deteri);
  endif;
  res_0=y - xl*phi_0;
else;
  res_0=res;
endif;
sigma=madc(res_0);

if not (ifail==-1);
/* compute psi_function, SSR_0 and matrix with regressor weigths */
psi_0=weightse.*(res_0./(sigma.*(weightsx+(1e-5.*(weightsx.<1e-5)))));
SSR_0=psi_0'psi_0;

/* compute LM_4 and corresponding pvalue */
{res_1,LM_4,pval_4}=
   CMPLMT(psi_0,k_l,weightsx.*aux_m,k_nlnl+4*k_nl-k_tl-3*k_tnlc);

/* compute LM_3 and corresponding pvalue */
{res_1,LM_3,pval_3}=
   CMPLMT(psi_0,k_l,weightsx.*aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc],
           k_nlnl+3*k_nl-k_tl-2*k_tnlc);

/* compute LM_H1 and corresponding pvalue */
{res_1,LM_H1,pval_H1}=
  CMPLMT(psi_0,k_l,weightsx.*aux_m[.,1:k_l+k_nlnl+k_nl-k_tl],k_nlnl+k_nl-k_tl);
endif;

/* compute LM_HL and corresponding pvalue */
logi=seqa(1,1,k_l+k_nlnl)|seqa(k_l+k_nlnl+k_nl-k_tl+1,1,k_nl-k_tnlc)|
     seqa(k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc+1,1,k_nl-k_tnlc);

/* compute sum of squared residuals under the null hypothesis */
phi_0=LMS(y,aux_m[.,logi]);
if (not GM);
  {phi_0,weightse,ifail}=M_EST(y,aux_m[.,logi],phi_0,&psi_pol,maxit,tol);
  weightsx=1;
else;
  {phi_0,weightsx,weightse,ifail}=
      GM_EST(y,aux_m[.,logi],phi_0,&psi_polx,&psi_pol,maxit,tol,deteri);
endif;
res_0=y - aux_m[.,logi]*phi_0;
sigma=madc(res_0);

if not (ifail==-1);
/* compute psi_function, SSR_0 and matrix with regressor weigths */
psi_0=weightse.*(res_0./(sigma.*(weightsx+(1e-5.*(weightsx.<1e-5)))));
SSR_0=psi_0'psi_0;

{res_1,LM_HL,pval_HL}=CMPLMT(psi_0,k_l+k_nlnl+2*k_nl-2*k_tnlc,
        weightsx.*aux_m,2*k_nl-k_tl-k_tnlc);
endif;

/* compute LM_HE and corresponding pvalue */
expi=seqa(1,1,k_l+k_nlnl+k_nl-k_tl)|
     seqa(k_l+k_nlnl+2*k_nl-k_tl-k_tnlc+1,1,k_nl-k_tnlc);

/* compute sum of squared residuals under the null hypothesis */
phi_0=LMS(y,aux_m[.,expi]);
if (not GM);
  {phi_0,weightse,ifail}=M_EST(y,aux_m[.,expi],phi_0,&psi_pol,maxit,tol);
  weightsx=1;
else;
  {phi_0,weightsx,weightse,ifail}=
      GM_EST(y,aux_m[.,expi],phi_0,&psi_polx,&psi_pol,maxit,tol,deteri);
endif;
res_0=y - aux_m[.,expi]*phi_0;
sigma=madc(res_0);

if not (ifail==-1);
/* compute psi_function, SSR_0 and matrix with regressor weigths */
psi_0=weightse.*(res_0./(sigma.*(weightsx+(1e-5.*(weightsx.<1e-5)))));
SSR_0=psi_0'psi_0;

{res_1,LM_HE,pval_HE}=CMPLMT(psi_0,k_l+k_nlnl+2*k_nl-k_tl-k_tnlc,
    weightsx.*aux_m,2*k_nl-2*k_tnlc);
endif;

/* compute LM_H4 and corresponding pvalue */
/* compute sum of squared residuals under the null hypothesis */
phi_0=LMS(y,aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc]);
if (not GM);
  {phi_0,weightse,ifail}=M_EST(y,aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc],
    phi_0,&psi_pol,maxit,tol);
  weightsx=1;
else;
  {phi_0,weightsx,weightse,ifail}=
      GM_EST(y,aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc],
        phi_0,&psi_polx,&psi_pol,maxit,tol,deteri);
endif;
res_0=y - aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc]*phi_0;
sigma=madc(res_0);

if not (ifail==-1);
/* compute psi_function, SSR_0 and matrix with regressor weigths */
psi_0=weightse.*(res_0./(sigma.*(weightsx+(1e-5.*(weightsx.<1e-5)))));
SSR_0=psi_0'psi_0;

{res_1,LM_H4,pval_H4}=
  CMPLMT(psi_0,k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc,
    weightsx.*aux_m[.,1:k_l+k_nlnl+4*k_nl-k_tl-3*k_tnlc],k_nl-k_tnlc);
endif;

/* compute LM_H3 and corresponding pvalue */
/* compute sum of squared residuals under the null hypothesis */
phi_0=LMS(y,aux_m[.,1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc]);
if (not GM);
  {phi_0,weightse,ifail}=M_EST(y,aux_m[.,1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc],
    phi_0,&psi_pol,maxit,tol);
  weightsx=1;
else;
  {phi_0,weightsx,weightse,ifail}=
      GM_EST(y,aux_m[.,1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc],
        phi_0,&psi_polx,&psi_pol,maxit,tol,deteri);
endif;
res_0=y - aux_m[.,1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc]*phi_0;
sigma=madc(res_0);

if not (ifail==-1);
/* compute psi_function, SSR_0 and matrix with regressor weigths */
psi_0=weightse.*(res_0./(sigma.*(weightsx+(1e-5.*(weightsx.<1e-5)))));
SSR_0=psi_0'psi_0;

{res_1,LM_H3,pval_H3}=
  CMPLMT(psi_0,k_l+k_nlnl+2*k_nl-k_tl-k_tnlc,
   weightsx.*aux_m[.,1:k_l+k_nlnl+3*k_nl-k_tl-2*k_tnlc],k_nl-k_tnlc);
endif;

/* compute LM_H2 and corresponding pvalue */
/* compute sum of squared residuals under the null hypothesis */
phi_0=LMS(y,aux_m[.,1:k_l+k_nlnl+k_nl-k_tl]);
if (not GM);
  {phi_0,weightse,ifail}=M_EST(y,aux_m[.,1:k_l+k_nlnl+k_nl-k_tl],
    phi_0,&psi_pol,maxit,tol);
  weightsx=1;
else;
  {phi_0,weightsx,weightse,ifail}=
      GM_EST(y,aux_m[.,1:k_l+k_nlnl+k_nl-k_tl],
        phi_0,&psi_polx,&psi_pol,maxit,tol,deteri);
endif;
res_0=y - aux_m[.,1:k_l+k_nlnl+k_nl-k_tl]*phi_0;
sigma=madc(res_0);

if not (ifail==-1);
/* compute psi_function, SSR_0 and matrix with regressor weigths */
psi_0=weightse.*(res_0./(sigma.*(weightsx+(1e-5.*(weightsx.<1e-5)))));
SSR_0=psi_0'psi_0;

{res_1,LM_H2,pval_H2}=
  CMPLMT(psi_0,k_l+k_nlnl+k_nl-k_tl,
    weightsx.*aux_m[.,1:k_l+k_nlnl+2*k_nl-k_tl-k_tnlc],k_nl-k_tnlc);
endif;

retp(LM_3,pval_3,LM_4,pval_4,LM_H4,pval_H4,LM_H3,pval_H3,
     LM_H2,pval_H2,LM_H1,pval_H1,LM_HL,pval_HL,LM_HE,pval_HE,weightse);
endp;


/* procedure to determine auxiliary regressors (how to handle constants, etc) */
proc(11)=DETAUXP(xli,cli,detli,k_l,xnli,cnli,detnli,k_nl,xtri,exc);
  local i,trl,k_tl,trnl,k_tnl,k_tnlc,nll,xnlnli,k_nlnl,temp;

/*
** Reorder linear and nonlinear regressors such that constant
** appears in first column
*/
if (cli==k_l) and (k_l GT 1);
  xli=xli[k_l 1:k_l-1];
  cli=1;
  if (detli /=0);
    detli=detli+1;
  endif;
elseif (cli GT 1);
  xli=xli[cli 1:cli-1 cli+1:k_l];
  cli=1;
  if (detli /=0);
    detli=detli+(detli .LT cli);
  endif;
endif;
if (detli /=0) and cli;
  detli=1|detli;
else;
  detli=1;
endif;
if (cnli==k_nl) and (k_nl GT 1);
  xnli=xnli[k_nl 1:k_nl-1];
  cnli=1;
  if (detnli /=0);
    detnli=detnli+1;
  endif;
elseif (cnli GT 1);
  xnli=xnli[cnli 1:cnli-1 cnli+1:k_nl];
  cnli=1;
  if (detnli /=0);
    detnli=detnli+(detnli .LT cnli);
  endif;
endif;

/*
** k_tl[i]=1 if i-th transition variable included as regressor in 
** linear part and nonlinear part contains constant 
*/
k_tl=zeros(rows(xtri),1);
i=0;
do until (i==rows(xtri));
  i=i+1;
  trl=sumc(xtri[i].==xli);
  if (trl and cnli) or exc[i];
    k_tl[i]=1;
  else;
    k_tl[i]=0;
  endif;
endo;

/*
** k_tnl[i]=1 if i-th transition variable included as regressor in 
** nonlinear part
** k_tnlc=1 if in addition nonlinear part contains constant 
*/
k_tnl=zeros(rows(xtri),1);
i=0;
do until (i==rows(xtri));
  i=i+1;  
  trnl=sumc(xtri[i].==xnli);
  if exc[i];
    k_tnl[i]=1;
    if (cnli AND (k_nl GT 1)); k_tnlc=1; else; k_tnlc=0; endif;
  elseif trnl;
    k_tnl[i]=1;
    if cnli;
      k_tnlc=1;
    else;
      k_tnlc=0;
    endif;
  else;
    k_tnl[i]=0;
    k_tnlc=0;
  endif;
endo;

/* Determine number of and which regressors in nonlinear part
    (other than transition variable if cnli) are not in linear part */
nll=sumc(xnli'.==xli);
xnlnli=delif(xnli,nll);
if cnli;
  i=0;
  do until (i==rows(xtri[i]));
    i=i+1;
    xnlnli=delif(xnlnli,xnlnli .==xtri[i]);
  endo;
endif;
if not scalmiss(xnlnli);
  k_nlnl=rows(xnlnli);
  if (detnli /=0);
    temp=cumsumc(1-nll);
    detnli=delif(temp[detnli],nll[detnli]);
  endif;
else;
  k_nlnl=0;
  detnli=0;
endif;

retp(xli,cli,xnli,cnli,k_tl,k_tnl,k_tnlc,k_nlnl,xnlnli,detli,detnli);
endp;


proc(3)=CMPLMT(res_0,k_0,aux_m,k_restr); /* compute standard LM statistic */
  local stdv,stdvzero,cind,T,LM_st,LM_pv,SSR_0,invaux_m,phi_1,res_1,SSR_1;

stdv=stdc(aux_m);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(aux_m)),stdvzero);
aux_m[.,cind]=(aux_m[.,cind]-meanc(aux_m[.,cind])')./(stdv[cind]');

T=rows(res_0);
res_1=1;
LM_st=zeros(2,1);
LM_pv=zeros(2,1);
SSR_0=res_0'res_0;
invaux_m=invpd(moment(aux_m,0));
if not scalmiss(invaux_m);
  phi_1=invaux_m*(aux_m'res_0);
  res_1=res_0 - aux_m*phi_1;
  SSR_1=res_1'res_1;

  LM_st[1]=T*(SSR_0 - SSR_1)/SSR_0;
  LM_pv[1]=CDFCHIC(LM_st[1],k_restr);
  LM_st[2]=((SSR_0 - SSR_1)/k_restr)/(SSR_1/(T-k_restr-k_0));
  LM_pv[2]=CDFFC(LM_st[2],k_restr,T-k_restr-k_0);
endif;

retp(res_1,LM_st,LM_pv);
endp;


proc(3)=CMPLMTI(res_0,k_0,aux_m,k_restr,SSR_1); /* compute standard LM statistic */
  local stdv,stdvzero,cind,T,LM_st,LM_pv,SSR_0,invaux_m,phi_1,res_1;

stdv=stdc(aux_m);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(aux_m)),stdvzero);
aux_m[.,cind]=(aux_m[.,cind]-meanc(aux_m[.,cind])')./(stdv[cind]');

T=rows(res_0);
res_1=1;
LM_st=zeros(2,1);
LM_pv=zeros(2,1);
invaux_m=invpd(moment(aux_m,0));
if not scalmiss(invaux_m);
  phi_1=invaux_m*(aux_m'res_0);
  res_1=res_0 - aux_m*phi_1;
  SSR_0=res_1'res_1;

  LM_st[1]=T*(SSR_0 - SSR_1)/SSR_0;
  LM_pv[1]=CDFCHIC(LM_st[1],k_restr);
  LM_st[2]=((SSR_0 - SSR_1)/k_restr)/(SSR_1/(T-k_restr-k_0));
  LM_pv[2]=CDFFC(LM_st[2],k_restr,T-k_restr-k_0);
endif;

retp(res_1,LM_st,LM_pv);
endp;


proc(2)=CMPLMTW(y,x_0,x_restr); /* compute Wooldridge-type LM statistic */
  local stdv,stdvzero,cind,T,k_0,k_restr,c,LM_st,LM_pv,inv_m,phi_0,res_0,phi_1,res_1,rr;

stdv=stdc(x_0);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(x_0)),stdvzero);
x_0[.,cind]=(x_0[.,cind]-meanc(x_0[.,cind])')./(stdv[cind]');

stdv=stdc(x_restr);
stdvzero=(stdv.==0);
cind=delif(seqa(1,1,cols(x_restr)),stdvzero);
x_restr[.,cind]=(x_restr[.,cind]-meanc(x_restr[.,cind])')./(stdv[cind]');

T=rows(y);
k_0=cols(x_0);
k_restr=cols(x_restr);

c=ones(T,1);
LM_st=zeros(2,1);
LM_pv=zeros(2,1);

inv_m=invpd(moment(x_0,0));
if not scalmiss(inv_m);
  phi_0=inv_m*(x_0'y);
  res_0=y - x_0*phi_0;
  phi_1=inv_m*(x_0'x_restr);
  res_1=x_restr - x_0*phi_1;
  rr=res_0.*res_1;
  LM_st[1]=(c'rr*invpd(rr'rr)*rr'c);
  LM_pv[1]=CDFCHIC(LM_st[1],k_restr);
  LM_st[2]=(LM_st[1]/k_restr)/((T-LM_st[1])/(T-k_restr-k_0));
  LM_pv[2]=CDFFC(LM_st[2],k_restr,T-k_restr-k_0);
endif;

retp(LM_st,LM_pv);
endp;
